/* ceelog expression compilation and evaluation.

Copyright (C) 2012 Red Hat, Inc. All rights reserved.
This copyrighted material is made available to anyone wishing to use, modify,
copy, or redistribute it subject to the terms and conditions of the GNU General
Public License v.2.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.

Author: Miloslav Trmač <mitr@redhat.com> */
#include <config.h>

#include <stdbool.h>
#include <string.h>

#include <glib.h>
#include <json/json_object.h>
#include <pcre.h>

#include <ceelog/event.h>
#include <ceelog/expression.h>

#include "event_private.h"
#include "expression_private.h"

/* We could just recursively walk the ceelog_expr tree, but we can do better
   with a little preparation.

   For example, for (c1 && c2 && (c3 || !c4) && c5 && c6), if c1 is false,
   we can immediately terminate; if c1...c4) is true and c5 is false, we can
   immediately terminate (ideally without returning from 5 nested function
   calls.

   So ceelog_expr is compiled into a sequence of insns, in this case:
   _0: goto c1 ? _1 : false;
   _1: goto c2 ? _2 : false;
   _2: goto c3 ? _4 : _3;
   _3: goto c4 ? false : _4;
   _4: goto c5 ? _5: false;
   _5: goto c6 ? true : false;

   There are several possible optimizations, e.g. only looking up a field value
   once if it is involved in several comparisons, doing the "cheap" comparisons
   first, etc. */

/* An instruction in expression evaluation */
struct insn
{
  enum internal_op op;
  size_t next_insn[2];	 /* Next insn index, depending on result of this one */
  /* For IO_FIELD_*.  Individual elements in the path, for
     g_ptr_array_free () */
  GPtrArray *field_path; /* For IO_FIELD_* */
  union
  {
    char *string;		/* IO_FIELD_{EQ,NE}; for g_free () */
    struct ceelog_regex re;	/* IO_*_{MATCHES,DOESNT_MATCH} */
  } v;
};

/* A "compiled" ceelog expression, ready for evaluation. */
struct ceelog_compiled_expr
{
  GArray *insns; 		/* Of "struct insn" */
  /* In a compiled expression, offset INSNS->LEN stands for "result is false",
     offset insns->LEN + 1 stands for "result is true". */
};

/* Free a compiled EXPR. */
void
ceelog_compiled_expr_free (struct ceelog_compiled_expr *expr)
{
  size_t i;

  for (i = 0; i < expr->insns->len; i++)
    {
      struct insn *insn;

      insn = &g_array_index (expr->insns, struct insn, i);
      switch (insn->op)
	{
	case IO_FIELD_EQ: case IO_FIELD_NE:
	  g_ptr_array_free (insn->field_path, TRUE);
	  g_free (insn->v.string);
	  break;

	case IO_FIELD_MATCHES: case IO_FIELD_DOESNT_MATCH:
	  g_ptr_array_free (insn->field_path, TRUE);
	  /* Fall through */
	case IO_UNSTRUCTURED_MATCHES: case IO_UNSTRUCTURED_DOESNT_MATCH:
	  ceelog_regex_free (&insn->v.re);
	  break;

	case IO_AND: case IO_OR: case IO_NOT: default:
	  g_return_if_reached ();
	}
    }

  g_array_free (expr->insns, TRUE);
  g_free (expr);
}

 /* Expression compilation */

/* A pending label fixup.
   During the compilation of a subexpression, we don't know the final
   values of next_insn yet.  Each such case is added to a "fixup" list that is
   interpreted when the next_insn value is determined. */
struct fixup
{
  size_t index;			/* Index of the insn to fix up */
  bool case_;			/* Which of the next_insn cases to fix up */
};

/* Allocate a new array of fixups */
static GArray *
new_fixups (void)
{
  return g_array_new (FALSE, FALSE, sizeof (struct fixup));
}

/* Set next_insn fields for FIXUPS in DEST to VALUE and free them */
static void
consume_fixups (struct ceelog_compiled_expr *dest, GArray *fixups, size_t value)
{
  size_t i;

  for (i = 0; i < fixups->len; i++)
    {
      struct fixup *f;
      struct insn *insn;

      f = &g_array_index (fixups, struct fixup, i);
      insn = &g_array_index (dest->insns, struct insn, f->index);
      insn->next_insn[f->case_] = value;
    }
  g_array_free (fixups, TRUE);
}

/* Parse and compile PATH.
   Return a compiled variant. */
static GPtrArray *
compile_field_path (const char *path)
{
  GPtrArray *res;
  const char *p;

  res = g_ptr_array_new_with_free_func (g_free);

  while ((p = strchr (path, '!')) != NULL)
    {
      g_ptr_array_add (res, g_strndup (path, p - path));
      path = p + 1;
    }
  g_ptr_array_add (res, g_strdup (path));

  return res;
}

/* Compile SUB into DEST.
   Return 0 if OK, -1 on error.
   Append fixups for the case of SUB being false to FALSE_FIXUPS, for the case
   of SUB being true to TRUE_FIXUPS.
   On error, keep SUB in a valid enough state so that
   ceelog_compiled_expr_free () works. */
static int
compile_sub (struct ceelog_compiled_expr *dest, struct ceelog_expr *sub,
	     GArray *false_fixups, GArray *true_fixups, GError **error)
{
  size_t i;
  struct insn *insn;
  struct fixup f;

  /* First, handle control structures (and return) */
  switch (sub->op)
    {
    case IO_AND:
      {
	GArray *fixups;

	fixups = new_fixups ();
	/* A && B: if A is false, return false */
	if (compile_sub (dest, sub->v.sub[0], false_fixups, fixups, error) != 0)
	  {
	    g_array_free (fixups, TRUE);
	    return -1;
	  }
	consume_fixups (dest, fixups, dest->insns->len);
	/* Otherwise return what B does */
	return compile_sub (dest, sub->v.sub[1], false_fixups, true_fixups,
			    error);
      }

    case IO_OR:
      {
	GArray *fixups;

	fixups = new_fixups ();
	/* A || B: if A is true, return true */
	if (compile_sub (dest, sub->v.sub[0], fixups, true_fixups, error) != 0)
	  {
	    g_array_free (fixups, TRUE);
	    return -1;
	  }
	consume_fixups (dest, fixups, dest->insns->len);
	/* Otherwise return what B does */
	return compile_sub (dest, sub->v.sub[1], false_fixups, true_fixups,
			    error);
      }

    case IO_NOT:
      return compile_sub (dest, sub->v.sub[0], true_fixups, false_fixups,
			  error);

    default:
      break; /* Continue below */
    }

  i = dest->insns->len;
  g_array_set_size (dest->insns, i + 1);
  insn = &g_array_index (dest->insns, struct insn, i);
  insn->op = sub->op;
  switch (insn->op)
    {
    case IO_FIELD_EQ: case IO_FIELD_NE:
      insn->field_path = compile_field_path (sub->v.cmp.field_path);
      insn->v.string = g_strdup (sub->v.cmp.value);
      break;

    case IO_FIELD_MATCHES: case IO_FIELD_DOESNT_MATCH:
      insn->field_path = compile_field_path (sub->v.cmp.field_path);
      if (ceelog_regex_compile (&insn->v.re, sub->v.cmp.value, error) != 0)
	{
	  g_ptr_array_free (insn->field_path, TRUE);
	  g_array_set_size (dest->insns, i);
	  return -1;
	}
      break;

    case IO_UNSTRUCTURED_MATCHES: case IO_UNSTRUCTURED_DOESNT_MATCH:
      if (ceelog_regex_compile (&insn->v.re, sub->v.cmp.value, error) != 0)
	{
	  g_array_set_size (dest->insns, i);
	  return -1;
	}
      break;

    case IO_AND: case IO_OR: case IO_NOT: default:
      g_return_val_if_reached (-1);
    }

  f.index = i;
  f.case_ = false;
  g_array_append_val (false_fixups, f);
  f.case_ = true;
  g_array_append_val (true_fixups, f);
  return 0;
}

/* Compile EXPR.
   Return the compiled expression, or NULL on error. */
struct ceelog_compiled_expr *
ceelog_compile_expr (struct ceelog_expr *expr, GError **error)
{
  struct ceelog_compiled_expr *ce;
  GArray *false_fixups, *true_fixups;

  g_return_val_if_fail (expr != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  ce = g_new (struct ceelog_compiled_expr, 1);
  ce->insns = g_array_new (FALSE, FALSE, sizeof (struct insn));

  false_fixups = new_fixups ();
  true_fixups = new_fixups ();
  if (compile_sub (ce, expr, false_fixups, true_fixups, error) != 0)
    {
      g_array_free (false_fixups, TRUE);
      g_array_free (true_fixups, TRUE);
      ceelog_compiled_expr_free (ce);
    }
  consume_fixups (ce, false_fixups, ce->insns->len);
  consume_fixups (ce, true_fixups, ce->insns->len + 1);

  return ce;
}

/* Parse and compile text EXPR.
   Return the compiled expression, or NULL on error. */
struct ceelog_compiled_expr *
ceelog_parse_and_compile_expr (const char *expr, GError **error)
{
  struct ceelog_expr *e;
  struct ceelog_compiled_expr *res;

  e = ceelog_expr_parse (expr, error);
  if (e == NULL)
    return NULL;

  res = ceelog_compile_expr (e, error);
  ceelog_expr_free (e);
  return res;
}

 /* Expression evaluation */

/* Evaluation state */
struct eval_state
{
  struct ceelog_event *event;
  struct json_object *json;	/* For json_object_put (), or NULL */
  bool got_json;
};

/* Get field corresponding to INSN->field_path in S.
   Return the value, or NULL if not present or on error. */
static const char *
eval_field_path (struct eval_state *s, struct insn *insn)
{
  struct json_object *obj;
  size_t i;

  if (!s->got_json)
    {
      s->json = ceelog_event_get_json (s->event);
      s->got_json = true;
    }

  obj = s->json;
  for (i = 0; i < insn->field_path->len; i++)
    {
      if (obj == NULL)
	return NULL;

      obj = json_object_object_get (obj,
				    g_ptr_array_index (insn->field_path, i));
    }
  if (obj == NULL)
    return NULL;

  /* We could use json_object_get_string_len() to avoid a strlen() if we
     restricted obj to a json_type_string. */
  return json_object_get_string (obj);
}

/* Evaluate INSN against S.
   Return result of the evaluation, or false on error. */
static bool
eval_insn (struct eval_state *s, struct insn *insn)
{
  const char *v;

  switch (insn->op)
    {
    case IO_FIELD_EQ:
      v = eval_field_path (s, insn);
      return v != NULL && strcmp (v, insn->v.string) == 0;

    case IO_FIELD_NE:
      v = eval_field_path (s, insn);
      return v != NULL && strcmp (v, insn->v.string) != 0;

    case IO_FIELD_MATCHES:
      v = eval_field_path (s, insn);
      return v != NULL && pcre_exec (insn->v.re.re, insn->v.re.extra, v,
				     strlen (v), 0, 0, NULL, 0) == 0;

    case IO_FIELD_DOESNT_MATCH:
      v = eval_field_path (s, insn);
      return v != NULL && pcre_exec (insn->v.re.re, insn->v.re.extra, v,
				     strlen (v), 0, 0, NULL, 0) != 0;

    case IO_UNSTRUCTURED_MATCHES:
      {
	size_t len;

	v = ceelog_event_get_message_pointer (s->event, &len);
	return v != NULL && pcre_exec (insn->v.re.re, insn->v.re.extra, v, len,
				       0, 0, NULL, 0) == 0;
      }

    case IO_UNSTRUCTURED_DOESNT_MATCH:
      {
	size_t len;

	v = ceelog_event_get_message_pointer (s->event, &len);
	return v != NULL && pcre_exec (insn->v.re.re, insn->v.re.extra, v, len,
				       0, 0, NULL, 0) != 0;
      }

    case IO_AND: case IO_OR: case IO_NOT: default:
      g_return_val_if_reached (false);
    }
}

/* Evaluate EXPR on EVENT, return a boolean result.
   On evaluation error, return false. */
bool
ceelog_eval_expr (struct ceelog_compiled_expr *expr, struct ceelog_event *event)
{
  struct eval_state s;
  size_t ip;

  g_return_val_if_fail (expr != NULL, false);
  g_return_val_if_fail (event != NULL, false);

  s.event = event;
  s.json = NULL;
  s.got_json = false;

  ip = 0;
  while (ip < expr->insns->len)
    {
      struct insn *insn;
      bool res;

      insn = &g_array_index (expr->insns, struct insn, ip);
      res = eval_insn (&s, insn);
      ip = insn->next_insn[res];
    }

  if (s.json != NULL)
    json_object_put (s.json);

  return ip > expr->insns->len;	/* == (ip == expr->insns->len + 1) */
}
